#!/usr/bin/env python3

import os, pytest, tempfile, datetime

from cozify import cloud, config, hub
from cozify.test import debug
from cozify.test.fixtures import *
from cozify.Error import AuthenticationError

## basic cloud.authenticate() tests


@pytest.mark.live
def test_cloud_authenticate():
    assert cloud.authenticate()


@pytest.mark.live
def test_cloud_authenticate_hub():
    assert cloud.authenticate(trustHub=False)


@pytest.mark.logic
def test_cloud_noninteractive_otp():
    with pytest.raises(AuthenticationError) as e_info:
        cloud._getotp()


## cloud.refresh() logic tests


@pytest.mark.logic
def test_cloud_refresh_cold(tmp_cloud):
    config.state.remove_option('Cloud', 'last_refresh')
    config.dump_state()
    assert cloud._need_refresh(force=False, expiry=tmp_cloud.expiry)


@pytest.mark.logic
def test_cloud_refresh_force(tmp_cloud):
    config.dump_state()
    assert cloud._need_refresh(force=True, expiry=datetime.timedelta(days=365))


@pytest.mark.logic
def test_cloud_refresh_invalid(tmp_cloud):
    config.state['Cloud']['last_refresh'] = 'intentionally bad'
    config.dump_state()
    assert cloud._need_refresh(force=False, expiry=datetime.timedelta(days=365))


@pytest.mark.logic
def test_cloud_refresh_expiry_over(tmp_cloud):
    config.dump_state()
    assert cloud._need_refresh(force=False, expiry=datetime.timedelta(hours=1))


@pytest.mark.logic
def test_cloud_refresh_expiry_not_over(tmp_cloud):
    config.dump_state()
    assert not cloud._need_refresh(force=False, expiry=datetime.timedelta(days=2))


## integration tests for remote


@pytest.mark.live
def test_cloud_remote_match(live_cloud, live_hub):
    config.dump_state()
    live_hub.ping()
    local_tz = live_hub.tz()
    remote_tz = live_hub.tz(remote=True)

    assert local_tz == remote_tz



#!/usr/bin/env python3

import os, pytest, tempfile, datetime
import hashlib, json

from absl import logging
from cozify import config, hub

from . import fixtures_devices as dev


@pytest.fixture
def tmp_cloud():
    obj = lambda: 0
    obj.configfile, obj.configpath = tempfile.mkstemp(suffix='tmp_cloud')
    obj.section = 'Cloud'
    obj.email = 'example@example.com'
    obj.token = 'eyJkb20iOiJ1ayIsImFsZyI6IkhTNTEyIiwidHlwIjoiSldUIn0.eyJyb2xlIjo4LCJpYXQiOjE1MTI5ODg5NjksImV4cCI6MTUxNTQwODc2OSwidXNlcl9pZCI6ImRlYWRiZWVmLWFhYWEtYmJiYi1jY2NjLWRkZGRkZGRkZGRkZCIsImtpZCI6ImRlYWRiZWVmLWRkZGQtY2NjYy1iYmJiLWFhYWFhYWFhYWFhYSIsImlzcyI6IkNsb3VkIn0.QVKKYyfTJPks_BXeKs23uvslkcGGQnBTKodA-UGjgHg'  # valid but useless jwt token.
    obj.expiry = datetime.timedelta(days=1)
    obj.now = datetime.datetime.now()
    obj.iso_now = obj.now.isoformat().split(".")[0]
    obj.yesterday = obj.now - datetime.timedelta(days=1)
    obj.iso_yesterday = obj.yesterday.isoformat().split(".")[0]
    config.setStatePath(obj.configpath)
    from cozify import cloud
    cloud._setAttr('email', obj.email)
    cloud._setAttr('remotetoken', obj.token)
    cloud._setAttr('last_refresh', obj.iso_yesterday)
    yield obj
    os.remove(obj.configpath)
    logging.error('exiting, tried to remove: {0}'.format(obj.configpath))


@pytest.fixture
def live_cloud():
    configfile, configpath = tempfile.mkstemp(suffix='live_cloud')
    config.setStatePath(configpath, copy_current=True)
    from cozify import cloud
    yield cloud
    config.setStatePath()
    os.remove(configpath)


@pytest.fixture
def tmp_hub(tmp_cloud):
    with Tmp_hub(tmp_cloud) as hub_obj:
        print('Tmp hub state for testing:')
        config.dump_state()
        yield hub_obj


@pytest.fixture()
def live_hub():
    config.setStatePath()  # default config assumed to be live
    print('Live hub state for testing:')
    config.dump_state()  # dump state so it's visible in failed test output
    from cozify import hub
    yield hub


@pytest.fixture()
def offline_device():
    dev = None
    store = None
    for i, d in hub.devices(capabilities=hub.capability.COLOR_HS).items():
        if not d['state']['reachable']:
            dev = i
            # store state so it can be restored after tests
            store = d['state']
            break
    if dev is None:
        logging.fatal('Cannot run device tests, no offline COLOR_HS device to test against.')

    yield dev
    hub.device_state_replace(dev, store)


@pytest.fixture()
def online_device():
    dev = None
    store = None
    devs = hub.devices(capabilities=hub.capability.BRIGHTNESS)
    for i, d in devs.items():
        if d['state']['reachable'] and 'test' in d['name']:
            dev = d
            # store state so it can be restored after tests
            store = d['state'].copy()
            break
    if dev is None:
        logging.error(
            'Cannot run certain device tests, no COLOR_HS device online where name includes \'test\'.'
        )
    logging.info('online_device state before use ({1}): {0}'.format(store, _h6_dict(store)))
    yield dev
    logging.info('online_device state after use, before rollback ({1}): {0}'.format(dev['state'], _h6_dict(dev['state'])))
    hub.device_state_replace(dev['id'], store)


class Tmp_hub():
    """Creates a temporary hub section (with test data) in a tmp_cloud
    """

    def __init__(self, tmp_cloud):
        self.id = 'deadbeef-aaaa-bbbb-cccc-tmphubdddddd'
        self.name = 'HubbyMcHubFace'
        self.host = '127.0.0.1'
        self.section = 'Hubs.{0}'.format(self.id)
        self.token = 'eyJkb20iOiJ1ayIsImFsZyI6IkhTNTEyIiwidHlwIjoiSldUIn0.eyJyb2xlIjo4LCJpYXQiOjE1MTI5ODg5NjksImV4cCI6MTUxNTQwODc2OSwidXNlcl9pZCI6ImRlYWRiZWVmLWFhYWEtYmJiYi1jY2NjLWRkZGRkZGRkZGRkZCIsImtpZCI6ImRlYWRiZWVmLWRkZGQtY2NjYy1iYmJiLWFhYWFhYWFhYWFhYSIsImlzcyI6IkNsb3VkIn0.QVKKYyfTJPks_BXeKs23uvslkcGGQnBTKodA-UGjgHg'  # valid but useless jwt token.

    @pytest.mark.usefixtures("tmp_cloud")
    def __enter__(self):
        config.state.add_section(self.section)
        config.state[self.section]['hubname'] = self.name
        config.state[self.section]['host'] = self.host
        config.state[self.section]['hubtoken'] = self.token
        config.state['Hubs']['default'] = self.id
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        config.setStatePath()

    def devices(self):
        return dev.device_ids, dev.devices

    def states(self):
        return dev.states

def _h6_dict(d):
    j = json.dumps(d)
    w = j.encode('utf8')
    h = hashlib.md5(w)
    return h.hexdigest()[:6]



#!/usr/bin/env python3
import pytest

from cozify import cloud, hub, hub_api, config
from cozify.test import debug
from cozify.test.fixtures import *


@pytest.mark.live
def test_hub(live_cloud, live_hub):
    assert live_hub.ping()
    hub_id = live_hub.default()
    assert hub_api.hub(
        hub_id=hub_id,
        host=live_hub.host(hub_id),
        remote=live_hub.remote(hub_id),
        cloud_token=live_cloud.token(),
        hub_token=live_hub.token(hub_id))



class APIError(Exception):
    """Error raised for non-200 API return codes

    Args:
        status_code(int): HTTP status code returned by the API
        message(str): Potential error message returned by the API
    
    Attributes:
        status_code(int): HTTP status code returned by the API
        message(str): Potential error message returned by the API
    """

    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return 'API error, %s: %s' % (self.status_code, self.message)


class AuthenticationError(Exception):
    """Error raised for nonrecoverable authentication failures.

    Args:
        message(str): Human readable error description
    
    Attributes:
        message(str): Human readable error description
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return 'Authentication error: %s' % self.message



"""Module for handling Cozify Cloud API 1:1 functions

Attributes:
    cloudBase(str): API endpoint including version

"""

import json, requests

from .Error import APIError, AuthenticationError

cloudBase = 'https://cloud2.cozify.fi/ui/0.2/'


def requestlogin(email):  # pragma: no cover
    """Raw Cloud API call, request OTP to be sent to account email address.

    Args:
        email(str): Email address connected to Cozify account.
    """

    payload = {'email': email}
    response = requests.post(cloudBase + 'user/requestlogin', params=payload, timeout=5)
    if response.status_code is not 200:
        raise APIError(response.status_code, response.text)


def emaillogin(email, otp):  # pragma: no cover
    """Raw Cloud API call, request cloud token with email address & OTP.

    Args:
        email(str): Email address connected to Cozify account.
        otp(int): One time passcode.

    Returns:
        str: cloud token
    """

    payload = {'email': email, 'password': otp}

    response = requests.post(cloudBase + 'user/emaillogin', params=payload, timeout=5)
    if response.status_code == 200:
        return response.text
    else:
        raise APIError(response.status_code, response.text)


def lan_ip():  # pragma: no cover
    """1:1 implementation of hub/lan_ip

    This call will fail with an APIError if the requesting source address is not the same as that of the hub, i.e. if they're not in the same NAT network.
    The above is based on observation and may only be partially true.

    Returns:
        list: List of Hub ip addresses.
    """
    response = requests.get(cloudBase + 'hub/lan_ip', timeout=5)
    if response.status_code == 200:
        return json.loads(response.text)
    else:
        raise APIError(response.status_code, response.text)


def hubkeys(cloud_token):  # pragma: no cover
    """1:1 implementation of user/hubkeys

    Args:
        cloud_token(str) Cloud remote authentication token.

    Returns:
        dict: Map of hub_id: hub_token pairs.
    """
    headers = {'Authorization': cloud_token}
    response = requests.get(cloudBase + 'user/hubkeys', headers=headers, timeout=5)
    if response.status_code == 200:
        return json.loads(response.text)
    else:
        raise APIError(response.status_code, response.text)


def refreshsession(cloud_token):  # pragma: no cover
    """1:1 implementation of user/refreshsession

    Args:
        cloud_token(str) Cloud remote authentication token.

    Returns:
        str: New cloud remote authentication token. Not automatically stored into state.
    """
    headers = {'Authorization': cloud_token}
    response = requests.get(cloudBase + 'user/refreshsession', headers=headers, timeout=5)
    if response.status_code == 200:
        return response.text
    else:
        raise APIError(response.status_code, response.text)


def remote(cloud_token, hub_token, apicall, payload=None, **kwargs):  # pragma: no cover
    """1:1 implementation of 'hub/remote'

    Args:
        cloud_token(str): Cloud remote authentication token.
        hub_token(str): Hub authentication token.
        apicall(str): Full API call that would normally go directly to hub, e.g. '/cc/1.6/hub/colors'
        payload(str): json string to use as payload, changes method to PUT.

    Returns:
        requests.response: Requests response object.
    """

    headers = {'Authorization': cloud_token, 'X-Hub-Key': hub_token}
    if payload:
        response = requests.put(
            cloudBase + 'hub/remote' + apicall, headers=headers, data=payload, timeout=5)
    else:
        response = requests.get(cloudBase + 'hub/remote' + apicall, headers=headers, timeout=5)

    return response



from setuptools import setup

import sys
import cozify

with open('README.rst') as file:
    long_description = file.read()

setup(
    name='cozify',
    version=cozify.__version__,
    python_requires='>=3.4',
    author='artanicus',
    author_email='python-cozify@nocturnal.fi',
    url='https://github.com/Artanicus/python-cozify',
    project_urls={"Documentation": "https://python-cozify.readthedocs.io/"},
    description='Unofficial Python3 client library for the Cozify API.',
    long_description=long_description,
    license='MIT',
    packages=['cozify'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    install_requires=['requests', 'absl-py'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ])  # yapf: disable



#!/usr/bin/env python3
import pytest

from cozify import hub, multisensor
from cozify.test import debug
from cozify.test.fixtures import *
from cozify.Error import APIError


@pytest.mark.logic
def test_hub_tmp_hub(tmp_hub):
    assert config.state['Cloud']['email'] == 'example@example.com'
    assert config.state['Hubs']['default'] == 'deadbeef-aaaa-bbbb-cccc-tmphubdddddd'


@pytest.mark.live
def test_hub_tz(live_hub):
    assert hub.ping()
    assert hub.tz()


@pytest.mark.live
def test_hub_remote_naive(live_hub):
    assert hub.tz()


@pytest.mark.logic
def test_hub_remote_set(tmp_hub):
    assert hub.remote(tmp_hub.id, True) == True
    assert hub.remote(tmp_hub.id) == True
    assert hub.remote(tmp_hub.id, False) == False
    assert hub.remote(tmp_hub.id) == False


@pytest.mark.logic
def test_hub_autoremote_set(tmp_hub):
    assert hub.autoremote(tmp_hub.id, True) == True
    assert hub.autoremote(tmp_hub.id) == True
    assert hub.autoremote(tmp_hub.id, False) == False
    assert hub.autoremote(tmp_hub.id) == False


@pytest.mark.logic
def test_hub_id_to_name(tmp_hub):
    assert hub.name(tmp_hub.id) == tmp_hub.name


@pytest.mark.logic
def test_hub_name_to_id(tmp_hub):
    assert hub.hub_id(tmp_hub.name) == tmp_hub.id


@pytest.mark.live
def test_multisensor(live_hub):
    assert hub.ping()
    data = hub.devices()
    print(multisensor.getMultisensorData(data))


@pytest.mark.logic
def test_hub_get_id(tmp_hub):
    assert hub._get_id(hub_id=tmp_hub.id) == tmp_hub.id
    assert hub._get_id(hub_name=tmp_hub.name) == tmp_hub.id
    assert hub._get_id(hub_name=tmp_hub.name, hub_id=tmp_hub.id) == tmp_hub.id
    assert hub._get_id(hubName=tmp_hub.name) == tmp_hub.id
    assert hub._get_id(hubId=tmp_hub.id) == tmp_hub.id
    assert hub._get_id() == tmp_hub.id
    assert not hub._get_id(hub_id='foo') == tmp_hub.id


@pytest.mark.destructive
def test_hub_ping_autorefresh(live_hub):
    assert hub.ping()
    hub_id = live_hub.default()
    live_hub.token(hub_id=hub_id, new_token='destroyed-on-purpose-by-destructive-test')

    assert not live_hub.ping(autorefresh=False)
    with pytest.raises(APIError):
        hub.tz()
    assert live_hub.ping(autorefresh=True)


@pytest.mark.live
def test_hub_fill_kwargs(live_hub):
    assert hub.ping()
    kwargs = {}
    hub._fill_kwargs(kwargs)
    for key in ['hub_id', 'remote', 'autoremote', 'hub_token', 'cloud_token', 'host']:
        assert key in kwargs, 'key {0} did not get set.'.format(key)
        if key != 'host' or (key == 'host' and not live_hub.remote(hub.default())):
            assert kwargs[key] is not None, 'key {0} was set to None'.format(key)


@pytest.mark.logic
def test_hub_clean_state(tmp_hub):
    states = tmp_hub.states()
    assert states['clean'] == hub._clean_state(states['dirty'])


@pytest.mark.logic
def test_hub_in_range():
    assert hub._in_range(0.5, low=0.0, high=1.0)
    assert hub._in_range(0.0, low=0.0, high=1.0)
    assert hub._in_range(1.0, low=0.0, high=1.0)
    assert hub._in_range(None, low=0.0, high=1.0)
    with pytest.raises(ValueError):
        hub._in_range(1.5, low=0.0, high=1.0)
    with pytest.raises(ValueError):
        hub._in_range(-0.5, low=0.0, high=1.0)


@pytest.mark.destructive
@pytest.mark.remote
def test_hub_dirty_remote(live_hub):
    # test if we are remote to get meaningful results
    live_hub.ping(live_hub.default())
    if not live_hub.remote(live_hub.default()):
        pytest.xfail("Not remote, cannot run this test")
    else:
        # fuck up the state on purpose to say we're not remote
        assert not live_hub.remote(live_hub.default(), False)
        # attempt to repair the state
        assert live_hub.ping()
        # verify we're now considered to be remote
        assert live_hub.remote(live_hub.default())



"""Module for handling Cozify Cloud highlevel operations.
"""

from absl import logging
import datetime

from . import config
from . import hub_api
from . import cloud_api

from .Error import APIError, AuthenticationError


def authenticate(trustCloud=True, trustHub=True, remote=False, autoremote=True):
    """Authenticate with the Cozify Cloud and Hub.

    Interactive only when absolutely needed, mostly on the first run.
    By default authentication is run selectively only for the portions needed.
    Hub authentication lives in the Cloud module since the authentication is obtained from
    the cloud.

    Authentication is a multistep process:
        - trigger sending OTP to email address
        - perform email login with OTP to acquire cloud token
        - acquire hub information and authenticate with hub with cloud token
        - store hub token for further use

    Args:
        trustCloud(bool): Trust current stored state of cloud auth. Default True.
        trustHub(bool): Trust current stored state of hub auth. Default True.
        remote(bool): Treat a hub as being outside the LAN, i.e. calls will be routed via the Cozify Cloud remote call system. Defaults to False.
        autoremote(bool): Autodetect hub LAN presence and flip to remote mode if needed. Defaults to True.

    Returns:
        bool: True on authentication success. Failure will result in an exception.
    """

    from . import hub

    if not _isAttr('email'):  # pragma: no cover
        _setAttr('email', _getEmail())
    email = _getAttr('email')

    if _need_cloud_token(trustCloud):  # pragma: no cover
        try:
            cloud_api.requestlogin(email)
        except APIError:
            resetState()  # a bogus email will shaft all future attempts, better to reset
            raise

        # get OTP from user, not stored anywhere since they have a very short lifetime
        otp = _getotp()
        try:
            cloud_token = cloud_api.emaillogin(email, otp)
        except APIError:
            logging.error('OTP authentication has failed.')
            resetState()
            raise

        # save the successful cloud_token
        _setAttr('last_refresh', config._iso_now(), commit=False)
        _setAttr('remoteToken', cloud_token, commit=True)
    else:
        # cloud_token already fine, let's just use it
        cloud_token = _getAttr('remoteToken')

    if _need_hub_token(trustHub):
        localHubs = cloud_api.lan_ip()  # will only work if we're local to the Hub, otherwise None
        # TODO(artanicus): unknown what will happen if there is a local hub but another one remote. Needs testing by someone with multiple hubs. Issue #7
        hubkeys = cloud_api.hubkeys(
            cloud_token)  # get all registered hubs and their keys from the cloud.
        if not hubkeys:
            logging.fatal(
                'You have not registered any hubs to the Cozify Cloud, hence a hub cannot be used yet.'
            )

        # evaluate all returned Hubs and store them
        for hub_id, hub_token in hubkeys.items():
            logging.debug('hub: {0} token: {1}'.format(hub_id, hub_token))
            hub_info = None
            hub_ip = None

            if not hub.exists(hub_id):
                autoremote = True
            else:
                autoremote = hub.autoremote(hub_id=hub_id)
            # if we're remote, we didn't get a valid ip
            if not localHubs:
                logging.info('No local Hubs detected, changing to remote mode.')
                hub_info = hub_api.hub(remote=True, cloud_token=cloud_token, hub_token=hub_token)
                # if the hub wants autoremote we flip the state. If this is the first time the hub is seen, act as if autoremote=True, remote=False
                if not hub.exists(hub_id) or (hub.autoremote(hub_id) and not hub.remote(hub_id)):
                    logging.info('[autoremote] Flipping hub remote status from local to remote.')
                    remote = True
            else:
                # localHubs is valid so a hub is in the lan. A mixed environment cannot yet be detected.
                # cloud_api.lan_ip cannot provide a map as to which ip is which hub. Thus we actually need to determine the right one.
                # TODO(artanicus): Need to truly test how multihub works before implementing ip to hub resolution. See issue #7
                hub_ip = localHubs[0]
                hub_info = hub_api.hub(host=hub_ip, remote=False)
                # if the hub wants autoremote we flip the state. If this is the first time the hub is seen, act as if autoremote=True, remote=False
                if not hub.exists(hub_id) or (hub.autoremote(hub_id) and hub.remote(hub_id)):
                    logging.info('[autoremote] Flipping hub remote status from remote to local.')
                    remote = False

            hub_name = hub_info['name']
            if hub_id in hubkeys:
                hub_token = hubkeys[hub_id]
            else:  # pragma: no cover
                logging.error('The hub "{0}" is not linked to the given account: "{1}"'.format(
                    hub_name, _getAttr('email')))
                resetState()
                return False

            # if hub name not already known, create named section
            hubSection = 'Hubs.' + hub_id
            if hubSection not in config.state:
                config.state.add_section(hubSection)
            # if default hub not set, set this hub as the first as the default
            if 'default' not in config.state['Hubs']:
                config.state['Hubs']['default'] = hub_id

            # store Hub data under it's named section
            hub._setAttr(hub_id, 'host', hub_ip, commit=False)
            hub._setAttr(hub_id, 'hubName', hub_name, commit=False)
            hub.token(hub_id, hub_token)
            hub.remote(hub_id, remote)
    return True


def resetState():
    """Reset stored cloud state.

    Any further authentication flow will start from a clean slate.
    Hub state is left intact.
    """

    config.state['Cloud'] = {}
    config.stateWrite()


def ping(autorefresh=True, expiry=None):
    """Test cloud token validity. On success will also trigger a refresh if it's needed by the current key expiry.

    Args:
        refresh(bool): Wether to perform a autorefresh check after a successful ping. Defaults to True.
        expiry(datetime.timedelta): timedelta object for duration how often cloud_token will be auto-refreshed when cloud.ping() is called. If not set, cloud.refresh() defaults are used.

    Returns:
        bool: validity of stored token.

    """

    try:
        cloud_api.hubkeys(token())  # TODO(artanicus): see if there's a cheaper API call
    except APIError as e:  # pragma: no cover
        if e.status_code == 401:
            return False
        else:
            raise
    else:
        if expiry:  # pragma: no cover
            refresh(expiry=expiry)
        else:  # let refresh use it's default expiry
            refresh()
        return True


def refresh(force=False, expiry=datetime.timedelta(days=1)):
    """Renew current cloud token and store new token in state.

    This call will only succeed if the current cloud token is still valid.
    A new refreshed token is requested from the API only if sufficient time has passed since the previous refresh.

    Args:
        force(bool): Set to True to always perform a refresh regardless of time passed since previous refresh.
        expiry(datetime.timedelta): timedelta object for duration of refresh expiry. Defaults to one day.

    Returns:
        bool: Success of refresh attempt, True also when expiry wasn't over yet even though no refresh was performed.
    """
    if _need_refresh(force, expiry):
        try:
            cloud_token = cloud_api.refreshsession(token())
        except APIError as e:  # pragma: no cover
            if e.status_code == 401:
                # too late, our token is already dead
                return False
            else:
                raise
        else:
            _setAttr('last_refresh', config._iso_now(), commit=False)
            token(cloud_token)
            logging.info('cloud_token has been successfully refreshed.')

            return True
    else:
        logging.debug(
            "Not refreshing token, it's not old enough yet. Limit is: {0})".format(expiry))


def _need_refresh(force, expiry):
    """Evaluate if refresh timer is already over or if forcing is valid.

    Args:
        force(bool): Set to True to always perform a refresh regardless of time passed since previous refresh.
        expiry(datetime.timedelta): timedelta object for duration of refresh expiry.

    Returns:
        bool: True if refresh should be done according to forcing and expiry.
    """

    last_refresh_str = None

    try:
        last_refresh_str = _getAttr('last_refresh')
    except AttributeError:  # not stored in state yet, e.g. first refresh
        logging.info("Last cloud token refresh unknown, will force refresh.")
        force = True
    else:
        try:
            last_refresh = datetime.datetime.strptime(last_refresh_str, "%Y-%m-%dT%H:%M:%S")
        except ValueError:  # not readable as a timestamp
            logging.error("Last cloud token refresh timestamp invalid, will force refresh.")
            force = True

    now = datetime.datetime.now()

    if force or last_refresh + expiry < now:
        return True


def _need_cloud_token(trust=True):
    """Validate current remote token and decide if we'll request it during authentication.

    Args:
        trust(bool): Set to False to always decide to renew. Defaults to True.

    Returns:
        bool: True to indicate a need to request token.
    """

    # check if we've got a cloud_token before doing expensive checks
    if trust and 'remoteToken' in config.state['Cloud']:
        if config.state['Cloud']['remoteToken'] is None:  # pragma: no cover
            return True
        else:  # perform more expensive check
            return not ping()
    return True


def _need_hub_token(trust=True):
    """Validate current hub token and decide if we'll request it during authentication.

    Args:
        trust(bool): Set to False to always decide to renew. Defaults to True.

    Returns:
        bool: True to indicate a need to request token.
    """
    from . import hub

    if not trust:
        logging.debug("hub_token not trusted so we'll say it needs to be renewed.")
        return True

    # First do quick checks, i.e. do we even have a token already
    if 'default' not in config.state['Hubs'] or 'hubtoken' not in config.state['Hubs.' +
                                                                               config.state['Hubs']
                                                                               ['default']]:
        logging.debug("We don't have a valid hubtoken or it's not trusted.")
        return True
    else:  # if we have a token, we need to test if the API is callable
        # avoid compliating things by disabling autorefresh on failure.
        ping = hub.ping(autorefresh=False)
        logging.debug("Testing hub.ping() for hub_token validity: {0}".format(ping))
        return not ping


def _getotp():
    try:
        return input('OTP from your email: ')
    except (EOFError, IOError):  # if running non-interactive or ^d
        message = "OTP unavailable, authentication cannot succeed. This may happen if running non-interactively (closed stdin)."
        raise AuthenticationError(message)


def _getEmail():  # pragma: no cover
    return input('Enter your Cozify account email address: ')


def _getAttr(attr):
    """Get cloud state attributes by attr name

    Args:
        attr(str): Name of cloud state attribute to retrieve
    Returns:
        str: Value of attribute or exception on failure
    """
    section = 'Cloud'
    if section in config.state and attr in config.state[section]:
        return config.state[section][attr]
    else:
        logging.warning('Cloud attribute {0} not found in state.'.format(attr))
        raise AttributeError


def _setAttr(attr, value, commit=True):
    """Set cloud state attributes by attr name

    Args:
        attr(str): Name of cloud state attribute to overwrite. Attribute will be created if it doesn't exist.
        value(str): Value to store
        commit(bool): True to commit state after set. Defaults to True.
    """
    section = 'Cloud'
    if section in config.state:
        if attr not in config.state[section]:
            logging.info(
                "Attribute {0} was not already in {1} state, new attribute created.".format(
                    attr, section))
        config.state[section][attr] = value
        if commit:
            config.stateWrite()
    else:  # pragma: no cover
        logging.warning('Section {0} not found in state.'.format(section))
        raise AttributeError


def _isAttr(attr):
    """Check validity of attribute by attr name.

    Returns:
        bool: True if attribute exists
    """
    return attr in config.state['Cloud'] and config.state['Cloud'][attr]


def token(new_token=None):
    """Get currently used cloud_token or set a new one.

    Returns:
        str: Cloud remote authentication token.
    """
    if new_token:
        _setAttr('remotetoken', new_token)
    return _getAttr('remotetoken')


def email(new_email=None):  # pragma: no cover
    """Get currently used cloud account email or set a new one.

    Returns:
        str: Cloud user account email address.
    """
    if new_email:
        _setAttr('email', new_email)
    return _getAttr('email')



"""Module for all Cozify Hub API 1:1 calls

Attributes:
    apiPath(str): Hub API endpoint path including version. Things may suddenly stop working if a software update increases the API version on the Hub. Incrementing this value until things work will get you by until a new version is published.
"""

import requests, json, logging

from cozify import cloud_api

from .Error import APIError
from requests.exceptions import RequestException

apiPath = '/cc/1.12'


def _getBase(host, port=8893):
    return 'http://{0}:{1}'.format(host, port)


def get(call, hub_token_header=True, base=apiPath, **kwargs):
    """GET method for calling hub API.

    Args:
        call(str): API path to call after apiPath, needs to include leading /.
        hub_token_header(bool): Set to False to omit hub_token usage in call headers.
        base(str): Base path to call from API instead of global apiPath. Defaults to apiPath.
        **host(str): ip address or hostname of hub.
        **hub_token(str): Hub authentication token.
        **remote(bool): If call is to be local or remote (bounced via cloud).
        **cloud_token(str): Cloud authentication token. Only needed if remote = True.
    """
    return _call(
        method=requests.get,
        call='{0}{1}'.format(base, call),
        hub_token_header=hub_token_header,
        **kwargs)


def put(call, payload, hub_token_header=True, base=apiPath, **kwargs):
    """PUT method for calling hub API. For rest of kwargs parameters see get()

    Args:
        call(str): API path to call after apiPath, needs to include leading /.
        payload(str): json string to push out as the payload.
        hub_token_header(bool): Set to False to omit hub_token usage in call headers.
        base(str): Base path to call from API instead of global apiPath. Defaults to apiPath.
    """
    return _call(
        method=requests.put,
        call='{0}{1}'.format(base, call),
        hub_token_header=hub_token_header,
        payload=payload,
        **kwargs)


def _call(*, call, method, hub_token_header, payload=None, **kwargs):
    """Backend for get & put

    Args:
        call(str): Full API path to call.
        method(function): requests.get|put function to use for call.
    """
    response = None
    headers = {}
    if hub_token_header:
        if 'hub_token' not in kwargs:
            raise AttributeError('Asked to do a call to the hub but no hub_token provided.')
        headers['Authorization'] = kwargs['hub_token']
    if payload is not None:
        headers['content-type'] = 'application/json'

    if kwargs['remote']:  # remote call
        if 'cloud_token' not in kwargs:
            raise AttributeError('Asked to do remote call but no cloud_token provided.')
        response = cloud_api.remote(apicall=call, payload=payload, **kwargs)
    else:  # local call
        if not kwargs['host']:
            raise AttributeError(
                'Local call but no hostname was provided. Either set keyword remote or host.')
        try:
            response = method(
                _getBase(host=kwargs['host']) + call, headers=headers, data=payload, timeout=5)
        except RequestException as e:  # pragma: no cover
            raise APIError('connection failure',
                           'issues connection to \'{0}\': {1}'.format(kwargs['host'], e))

    # evaluate response, wether it was remote or local
    if response.status_code == 200:
        return response.json()
    elif response.status_code == 410:
        raise APIError(response.status_code,
                       'API version outdated. Update python-cozify. %s - %s - %s' %
                       (response.reason, response.url, response.text))  # pragma: no cover
    else:
        raise APIError(response.status_code,
                       '%s - %s - %s' % (response.reason, response.url, response.text))


def hub(**kwargs):
    """1:1 implementation of /hub API call. For kwargs see cozify.hub_api.get()

    Returns:
        dict: Hub state dict.
    """
    return get('hub', base='/', hub_token_header=False, **kwargs)


def tz(**kwargs):
    """1:1 implementation of /hub/tz API call. For kwargs see cozify.hub_api.get()

    Returns:
        str: Timezone of the hub, for example: 'Europe/Helsinki'
    """
    return get('/hub/tz', **kwargs)


def devices(**kwargs):
    """1:1 implementation of /devices API call. For remaining kwargs see cozify.hub_api.get()

    Args:
        **mock_devices(dict): If defined, returned as-is as if that were the result we received.

    Returns:
        dict: Full live device state as returned by the API
    """
    if 'mock_devices' in kwargs:
        return kwargs['mock_devices']

    return get('/devices', **kwargs)


def devices_command(command, **kwargs):
    """1:1 implementation of /devices/command. For kwargs see cozify.hub_api.put()

    Args:
        command(dict): dictionary of type DeviceData containing the changes wanted. Will be converted to json.

    Returns:
        str: What ever the API replied or raises an APIEerror on failure.
    """
    command = json.dumps(command)
    logging.debug('command json to send: {0}'.format(command))
    return put('/devices/command', command, **kwargs)


def devices_command_generic(*, device_id, command=None, request_type, **kwargs):
    """Command helper for CMD type of actions.
    No checks are made wether the device supports the command or not. For kwargs see cozify.hub_api.put()

    Args:
        device_id(str): ID of the device to operate on.
        request_type(str): Type of CMD to run, e.g. CMD_DEVICE_OFF
        command(dict): Optional dictionary to override command sent. Defaults to None which is interpreted as { device_id, type }
    Returns:
        str: What ever the API replied or raises an APIError on failure.
    """
    if command is None:
        command = [{"id": device_id, "type": request_type}]
    return devices_command(command, **kwargs)


def devices_command_state(*, device_id, state, **kwargs):
    """Command helper for CMD type of actions.
    No checks are made wether the device supports the command or not. For kwargs see cozify.hub_api.put()

    Args:
        device_id(str): ID of the device to operate on.
        state(dict): New state dictionary containing changes.
    Returns:
        str: What ever the API replied or raises an APIError on failure.
    """
    command = [{"id": device_id, "type": 'CMD_DEVICE', "state": state}]
    return devices_command(command, **kwargs)


def devices_command_on(device_id, **kwargs):
    """Command helper for CMD_DEVICE_ON.

    Args:
        device_id(str): ID of the device to operate on.
    Returns:
        str: What ever the API replied or raises an APIError on failure.
    """
    return devices_command_generic(device_id=device_id, request_type='CMD_DEVICE_ON', **kwargs)


def devices_command_off(device_id, **kwargs):
    """Command helper for CMD_DEVICE_OFF.

    Args:
        device_id(str): ID of the device to operate on.
    Returns:
        str: What ever the API replied or raises an APIException on failure.
    """
    return devices_command_generic(device_id=device_id, request_type='CMD_DEVICE_OFF', **kwargs)
